@@ -1,10 +1,12 @@
+from typing import Any
+
 from sgqlc.operation import Operation
 
 from pycarlo.lib import schema
 
 
 class Query(Operation):
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args: Any, **kwargs: Any):
         """
         An MCD Query operation.
 
@@ -14,7 +16,7 @@ class Query(Operation):
 
 
 class Mutation(Operation):
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args: Any, **kwargs: Any):
         """
         An MCD Mutation operation.
 
