@@ -1,33 +1,38 @@
 from unittest import TestCase
-from unittest.mock import patch, Mock
+from unittest.mock import Mock, patch
 from urllib.parse import urljoin
 
 from box import Box
-from requests.exceptions import Timeout
+from requests.exceptions import HTTPError, Timeout
 from responses import activate, mock, post, registries
 
 from pycarlo.common.errors import GqlError, InvalidSessionError
 from pycarlo.common.retries import Backoff
-from pycarlo.common.settings import DEFAULT_MCD_API_ENDPOINT, DEFAULT_MCD_IGW_ENDPOINT
-from pycarlo.core import Session, Client, Query
+from pycarlo.common.settings import (
+    DEFAULT_MCD_API_ENDPOINT,
+    DEFAULT_MCD_IGW_ENDPOINT,
+    RequestReason,
+)
+from pycarlo.core import Client, Query, Session
 
 MCD_API_ENDPOINT = DEFAULT_MCD_API_ENDPOINT
 MCD_IGW_ENDPOINT = DEFAULT_MCD_IGW_ENDPOINT
 
-MOCK_MCD_ID = 'foo'
-MOCK_MCD_TOKEN = 'bar'
-MOCK_SESSION_ID = 'qux'
-MOCK_TRACE_ID = 'baz'
-MOCK_USER_ID = 'uid'
+MOCK_MCD_ID = "foo"
+MOCK_MCD_TOKEN = "bar"
+MOCK_SESSION_ID = "qux"
+MOCK_TRACE_ID = "baz"
+MOCK_USER_ID = "uid"
 MOCK_HEADERS = {
-    'x-mcd-id': MOCK_MCD_ID,
-    'x-mcd-session-id': MOCK_SESSION_ID,
-    'x-mcd-token': MOCK_MCD_TOKEN,
-    'x-mcd-trace-id': MOCK_TRACE_ID,
-    'user-id': MOCK_USER_ID
+    "x-mcd-id": MOCK_MCD_ID,
+    "x-mcd-session-id": MOCK_SESSION_ID,
+    "x-mcd-telemetry-reason": RequestReason.USER.value,
+    "x-mcd-token": MOCK_MCD_TOKEN,
+    "x-mcd-trace-id": MOCK_TRACE_ID,
+    "user-id": MOCK_USER_ID,
 }
 MOCK_TIMEOUT = 16
-MOCK_EMAIL = 'dresden@montecarlodata.com'
+MOCK_EMAIL = "dresden@montecarlodata.com"
 MOCK_GET_USER_QUERY = """
 query {
   getUser {
@@ -35,11 +40,12 @@ query {
   }
 }
 """
-MOCK_GOOD_RESPONSE = {'data': {'getUser': {'email': MOCK_EMAIL}}, 'errors': []}
+MOCK_GOOD_RESPONSE = {"data": {"getUser": {"email": MOCK_EMAIL}}, "errors": []}
 
 _AIRFLOW_SCOPE = "AirflowCallbacks"
 _AIRFLOW_CALLBACKS_PATH = "/airflow/callbacks"
 
+
 class MockBackoff(Backoff):
     def __init__(self, count: int):
         super(MockBackoff, self).__init__(start=0, maximum=0)
@@ -70,12 +76,12 @@ class ClientTest(TestCase):
         self.assertEqual(self._client.session_name, MOCK_SESSION_ID)
 
     def test_get_session_endpoint(self):
-        endpoint = 'test.com'
+        endpoint = "test.com"
         self._session.endpoint = endpoint
         self.assertEqual(self._client.session_endpoint, endpoint)
 
-    @patch('pycarlo.core.client.uuid')
-    def test_get_headers(self, mock_uuid):
+    @patch("pycarlo.core.client.uuid")
+    def test_get_headers(self, mock_uuid: Mock):
         self._session.id = MOCK_MCD_ID
         self._session.token = MOCK_MCD_TOKEN
         self._session.session_name = MOCK_SESSION_ID
@@ -85,71 +91,95 @@ class ClientTest(TestCase):
 
         self.assertEqual(self._client._get_headers(), MOCK_HEADERS)
 
-    @patch.object(Client, '_get_headers', lambda *args: MOCK_HEADERS)
+    @patch.object(Client, "_get_headers", lambda *args: MOCK_HEADERS)
     @activate(registry=registries.OrderedRegistry)
     def test_call(self):
         post(MCD_API_ENDPOINT, json=MOCK_GOOD_RESPONSE, status=200)
         self.assertEqual(
-            self._client(query=MOCK_GET_USER_QUERY, retry_backoff=self._backoff, timeout_in_seconds=MOCK_TIMEOUT),
-            Box(MOCK_GOOD_RESPONSE, camel_killer_box=True).data
+            self._client(
+                query=MOCK_GET_USER_QUERY,
+                retry_backoff=self._backoff,
+                timeout_in_seconds=MOCK_TIMEOUT,
+            ),
+            Box(MOCK_GOOD_RESPONSE, camel_killer_box=True).data,
         )
         mock.assert_call_count(MCD_API_ENDPOINT, 1)
 
-    @patch.object(Client, '_get_headers', lambda *args: MOCK_HEADERS)
+    @patch.object(Client, "_get_headers", lambda *args: MOCK_HEADERS)
     @activate(registry=registries.OrderedRegistry)
     def test_call_with_object(self):
         post(MCD_API_ENDPOINT, json=MOCK_GOOD_RESPONSE, status=200)
 
         query = Query()
-        query.get_user.__fields__('email')
+        query.get_user.__fields__("email")
 
-        response = self._client(query=query, retry_backoff=self._backoff, timeout_in_seconds=MOCK_TIMEOUT)
+        response = self._client(
+            query=query, retry_backoff=self._backoff, timeout_in_seconds=MOCK_TIMEOUT
+        )
         self.assertEqual(response.get_user.email, MOCK_EMAIL)
         mock.assert_call_count(MCD_API_ENDPOINT, 1)
 
-    @patch.object(Client, '_get_headers', lambda *args: MOCK_HEADERS)
+    @patch.object(Client, "_get_headers", lambda *args: MOCK_HEADERS)
     @activate(registry=registries.OrderedRegistry)
     def test_call_with_single_gql_error(self):
         post(
             MCD_API_ENDPOINT,
-            json={'data': None, 'errors': [{'message': 'Life is a journey. Time is a river. The door is ajar.'}]},
+            json={
+                "data": None,
+                "errors": [{"message": "Life is a journey. Time is a river. The door is ajar."}],
+            },
             status=200,
         )
 
         with self.assertRaises(GqlError) as err:
-            self._client(query=MOCK_GET_USER_QUERY, retry_backoff=self._backoff, timeout_in_seconds=MOCK_TIMEOUT)
-        self.assertEqual('Life is a journey. Time is a river. The door is ajar.', err.exception.message)
+            self._client(
+                query=MOCK_GET_USER_QUERY,
+                retry_backoff=self._backoff,
+                timeout_in_seconds=MOCK_TIMEOUT,
+            )
+        self.assertEqual(
+            "Life is a journey. Time is a river. The door is ajar.", err.exception.message
+        )
         mock.assert_call_count(MCD_API_ENDPOINT, 1)
 
-    @patch.object(Client, '_get_headers', lambda *args: MOCK_HEADERS)
+    @patch.object(Client, "_get_headers", lambda *args: MOCK_HEADERS)
     @activate(registry=registries.OrderedRegistry)
     def test_call_with_multiple_gql_errors(self):
         post(
             MCD_API_ENDPOINT,
-            json={'data': None, 'errors': [{'message': 'Life is a journey.'}, {'message': 'Time is a river.'}]},
+            json={
+                "data": None,
+                "errors": [{"message": "Life is a journey."}, {"message": "Time is a river."}],
+            },
             status=200,
         )
 
         with self.assertRaises(GqlError) as err:
-            self._client(query=MOCK_GET_USER_QUERY, retry_backoff=self._backoff, timeout_in_seconds=MOCK_TIMEOUT)
-        self.assertEqual('Life is a journey.\nTime is a river.', err.exception.message)
+            self._client(
+                query=MOCK_GET_USER_QUERY,
+                retry_backoff=self._backoff,
+                timeout_in_seconds=MOCK_TIMEOUT,
+            )
+        self.assertEqual("Life is a journey.\nTime is a river.", err.exception.message)
         mock.assert_call_count(MCD_API_ENDPOINT, 1)
 
-    @patch.object(Client, '_get_headers', lambda *args: MOCK_HEADERS)
+    @patch.object(Client, "_get_headers", lambda *args: MOCK_HEADERS)
     @activate(registry=registries.OrderedRegistry)
     def test_call_with_user_errors(self):
         post(
             MCD_API_ENDPOINT,
-            json={'message': 'Invalid field value', 'field': 'name'},
+            json={"message": "Invalid field value", "field": "name"},
             status=400,
-            headers={'X-Custom': 'blah blah'},
+            headers={"X-Custom": "blah blah"},
         )
 
         backoff = MockBackoff(2)
         with self.assertRaises(GqlError) as err:
-            self._client(query=MOCK_GET_USER_QUERY, retry_backoff=backoff, timeout_in_seconds=MOCK_TIMEOUT)
+            self._client(
+                query=MOCK_GET_USER_QUERY, retry_backoff=backoff, timeout_in_seconds=MOCK_TIMEOUT
+            )
         self.assertEqual(
-            {'Content-Type': 'application/json', 'X-Custom': 'blah blah'},
+            {"Content-Type": "application/json", "X-Custom": "blah blah"},
             err.exception.headers,
         )
         self.assertEqual(
@@ -157,7 +187,7 @@ class ClientTest(TestCase):
             err.exception.status_code,
         )
         self.assertEqual(
-            'Invalid field value',
+            "Invalid field value",
             err.exception.message,
         )
         self.assertEqual(
@@ -165,20 +195,24 @@ class ClientTest(TestCase):
             err.exception.body,
         )
         self.assertEqual(
-            f'400 Client Error: Bad Request for url: {MCD_API_ENDPOINT}',
+            f"400 Client Error: Bad Request for url: {MCD_API_ENDPOINT}",
             str(err.exception),
         )
         mock.assert_call_count(MCD_API_ENDPOINT, 1)
 
-    @patch.object(Client, '_get_headers', lambda *args: MOCK_HEADERS)
+    @patch.object(Client, "_get_headers", lambda *args: MOCK_HEADERS)
     @activate(registry=registries.OrderedRegistry)
     def test_call_with_network_errors_no_retries(self):
-        post(MCD_API_ENDPOINT, body='Oops', status=503)
+        post(MCD_API_ENDPOINT, body="Oops", status=503)
 
         with self.assertRaises(GqlError) as err:
-            self._client(query=MOCK_GET_USER_QUERY, retry_backoff=self._backoff, timeout_in_seconds=MOCK_TIMEOUT)
+            self._client(
+                query=MOCK_GET_USER_QUERY,
+                retry_backoff=self._backoff,
+                timeout_in_seconds=MOCK_TIMEOUT,
+            )
         self.assertEqual(
-            {'Content-Type': 'text/plain'},
+            {"Content-Type": "text/plain"},
             err.exception.headers,
         )
         self.assertEqual(
@@ -186,49 +220,53 @@ class ClientTest(TestCase):
             err.exception.status_code,
         )
         self.assertEqual(
-            'Oops',
+            "Oops",
             err.exception.message,
         )
         self.assertEqual(
-            'Oops',
+            "Oops",
             err.exception.body,
         )
         self.assertEqual(
-            f'503 Server Error: Service Unavailable for url: {MCD_API_ENDPOINT}',
+            f"503 Server Error: Service Unavailable for url: {MCD_API_ENDPOINT}",
             str(err.exception),
         )
         mock.assert_call_count(MCD_API_ENDPOINT, 1)
 
-    @patch.object(Client, '_get_headers', lambda *args: MOCK_HEADERS)
+    @patch.object(Client, "_get_headers", lambda *args: MOCK_HEADERS)
     @activate(registry=registries.OrderedRegistry)
     def test_call_with_network_errors_retried_with_success(self):
-        post(MCD_API_ENDPOINT, body='Oops', status=503)
-        post(MCD_API_ENDPOINT, body='Oops', status=503)
+        post(MCD_API_ENDPOINT, body="Oops", status=503)
+        post(MCD_API_ENDPOINT, body="Oops", status=503)
         post(MCD_API_ENDPOINT, json=MOCK_GOOD_RESPONSE, status=200)
 
         backoff = MockBackoff(2)
-        response = self._client(query=MOCK_GET_USER_QUERY, retry_backoff=backoff, timeout_in_seconds=MOCK_TIMEOUT)
+        response = self._client(
+            query=MOCK_GET_USER_QUERY, retry_backoff=backoff, timeout_in_seconds=MOCK_TIMEOUT
+        )
         self.assertEqual(
             MOCK_EMAIL,
-            response.get_user.email,
+            response.get_user.email,  # type: ignore
         )
         mock.assert_call_count(MCD_API_ENDPOINT, 3)
 
-    @patch.object(Client, '_get_headers', lambda *args: MOCK_HEADERS)
+    @patch.object(Client, "_get_headers", lambda *args: MOCK_HEADERS)
     @activate(registry=registries.OrderedRegistry)
     def test_call_with_network_errors_retried_to_exhaustion(self):
-        post(MCD_API_ENDPOINT, body='Oops', status=503)
-        post(MCD_API_ENDPOINT, body='Oops', status=503)
-        post(MCD_API_ENDPOINT, body='Oops', status=503)
-        post(MCD_API_ENDPOINT, body='Oops', status=503)
-        post(MCD_API_ENDPOINT, body='Oops', status=503)
-        post(MCD_API_ENDPOINT, body='Oops', status=503)
+        post(MCD_API_ENDPOINT, body="Oops", status=503)
+        post(MCD_API_ENDPOINT, body="Oops", status=503)
+        post(MCD_API_ENDPOINT, body="Oops", status=503)
+        post(MCD_API_ENDPOINT, body="Oops", status=503)
+        post(MCD_API_ENDPOINT, body="Oops", status=503)
+        post(MCD_API_ENDPOINT, body="Oops", status=503)
 
         backoff = MockBackoff(4)
         with self.assertRaises(GqlError) as err:
-            self._client(query=MOCK_GET_USER_QUERY, retry_backoff=backoff, timeout_in_seconds=MOCK_TIMEOUT)
+            self._client(
+                query=MOCK_GET_USER_QUERY, retry_backoff=backoff, timeout_in_seconds=MOCK_TIMEOUT
+            )
         self.assertEqual(
-            {'Content-Type': 'text/plain'},
+            {"Content-Type": "text/plain"},
             err.exception.headers,
         )
         self.assertEqual(
@@ -236,20 +274,20 @@ class ClientTest(TestCase):
             err.exception.status_code,
         )
         self.assertEqual(
-            'Oops',
+            "Oops",
             err.exception.message,
         )
         self.assertEqual(
-            'Oops',
+            "Oops",
             err.exception.body,
         )
         self.assertEqual(
-            f'503 Server Error: Service Unavailable for url: {MCD_API_ENDPOINT}',
+            f"503 Server Error: Service Unavailable for url: {MCD_API_ENDPOINT}",
             str(err.exception),
         )
         mock.assert_call_count(MCD_API_ENDPOINT, 5)
 
-    @patch.object(Client, '_get_headers', lambda *args: MOCK_HEADERS)
+    @patch.object(Client, "_get_headers", lambda *args: MOCK_HEADERS)
     @activate(registry=registries.OrderedRegistry)
     def test_call_with_timeout_retried_with_success(self):
         post(MCD_API_ENDPOINT, body=Timeout())
@@ -257,14 +295,146 @@ class ClientTest(TestCase):
         post(MCD_API_ENDPOINT, json=MOCK_GOOD_RESPONSE, status=200)
 
         backoff = MockBackoff(2)
-        response = self._client(query=MOCK_GET_USER_QUERY, retry_backoff=backoff, timeout_in_seconds=MOCK_TIMEOUT)
+        response = self._client(
+            query=MOCK_GET_USER_QUERY, retry_backoff=backoff, timeout_in_seconds=MOCK_TIMEOUT
+        )
         self.assertEqual(
             MOCK_EMAIL,
-            response.get_user.email,
+            response.get_user.email,  # type: ignore
         )
         mock.assert_call_count(MCD_API_ENDPOINT, 3)
 
-    @patch.object(Client, '_get_headers', lambda *args: MOCK_HEADERS)
+    @patch.object(Client, "_get_headers", lambda *args: MOCK_HEADERS)
+    @activate(registry=registries.OrderedRegistry)
+    def test_call_with_400_errors_not_retried(self):
+        def should_retry(exc: Exception) -> bool:
+            if isinstance(exc, HTTPError):
+                return exc.response.status_code != 400  # type: ignore
+
+            return False
+
+        self._session.scope = _AIRFLOW_SCOPE
+        post(
+            MCD_API_ENDPOINT,
+            json={"errors": [{"message": "Request timed out", "code": "REQUEST_TIMEOUT"}]},
+            status=400,
+            headers={"X-Custom": "blah blah"},
+        )
+
+        backoff = MockBackoff(2)
+
+        with self.assertRaises(HTTPError) as err:
+            self._client.make_request(
+                path=MCD_API_ENDPOINT, body={}, should_retry=should_retry, retry_backoff=backoff
+            )
+
+        self.assertEqual(400, err.exception.response.status_code)  # type: ignore
+        mock.assert_call_count(MCD_API_ENDPOINT, 1)
+
+    @patch.object(Client, "_get_headers", lambda *args: MOCK_HEADERS)
+    @activate(registry=registries.OrderedRegistry)
+    def test_call_with_400_errors_retried(self):
+        def should_retry(exc: Exception) -> bool:
+            if isinstance(exc, HTTPError):
+                return exc.response.status_code == 400  # type: ignore
+
+            return False
+
+        self._session.scope = _AIRFLOW_SCOPE
+        post(
+            MCD_API_ENDPOINT,
+            json={"errors": [{"message": "Request timed out", "code": "REQUEST_TIMEOUT"}]},
+            status=400,
+            headers={"X-Custom": "blah blah"},
+        )
+        post(
+            MCD_API_ENDPOINT,
+            json={"errors": [{"message": "Request timed out", "code": "REQUEST_TIMEOUT"}]},
+            status=400,
+            headers={"X-Custom": "blah blah"},
+        )
+        post(
+            MCD_API_ENDPOINT,
+            json={"errors": [{"message": "Request timed out", "code": "REQUEST_TIMEOUT"}]},
+            status=400,
+            headers={"X-Custom": "blah blah"},
+        )
+
+        backoff = MockBackoff(2)
+
+        with self.assertRaises(HTTPError) as err:
+            self._client.make_request(
+                path=MCD_API_ENDPOINT, body={}, should_retry=should_retry, retry_backoff=backoff
+            )
+
+        self.assertEqual(400, err.exception.response.status_code)  # type: ignore
+        mock.assert_call_count(MCD_API_ENDPOINT, 3)
+
+    @patch.object(Client, "_get_headers", lambda *args: MOCK_HEADERS)
+    @activate(registry=registries.OrderedRegistry)
+    def test_call_with_graphql_server_timeout_retried_to_exhaustion(self):
+        post(
+            MCD_API_ENDPOINT,
+            json={"errors": [{"message": "Request timed out", "code": "REQUEST_TIMEOUT"}]},
+            status=200,
+            headers={"X-Custom": "blah blah"},
+        )
+        post(
+            MCD_API_ENDPOINT,
+            json={"errors": [{"message": "Request timed out", "code": "REQUEST_TIMEOUT"}]},
+            status=200,
+            headers={"X-Custom": "blah blah"},
+        )
+
+        backoff = MockBackoff(1)
+        with self.assertRaises(GqlError) as err:
+            self._client(
+                query=MOCK_GET_USER_QUERY, retry_backoff=backoff, timeout_in_seconds=MOCK_TIMEOUT
+            )
+        self.assertEqual(
+            {},
+            err.exception.headers,
+        )
+        self.assertEqual(
+            200,
+            err.exception.status_code,
+        )
+        self.assertEqual(
+            "Request timed out",
+            err.exception.message,
+        )
+        self.assertEqual(
+            [{"code": "REQUEST_TIMEOUT", "message": "Request timed out"}],
+            err.exception.body,
+        )
+        self.assertEqual(
+            "Request timed out",
+            str(err.exception),
+        )
+        mock.assert_call_count(MCD_API_ENDPOINT, 2)
+
+    @patch.object(Client, "_get_headers", lambda *args: MOCK_HEADERS)
+    @activate(registry=registries.OrderedRegistry)
+    def test_call_with_graphql_server_timeout_retried_with_succes(self):
+        post(
+            MCD_API_ENDPOINT,
+            json={"errors": [{"message": "Request timed out", "code": "REQUEST_TIMEOUT"}]},
+            status=200,
+            headers={"X-Custom": "blah blah"},
+        )
+        post(MCD_API_ENDPOINT, json=MOCK_GOOD_RESPONSE, status=200)
+
+        backoff = MockBackoff(2)
+        response = self._client(
+            query=MOCK_GET_USER_QUERY, retry_backoff=backoff, timeout_in_seconds=MOCK_TIMEOUT
+        )
+        self.assertEqual(
+            MOCK_EMAIL,
+            response.get_user.email,  # type: ignore
+        )
+        mock.assert_call_count(MCD_API_ENDPOINT, 2)
+
+    @patch.object(Client, "_get_headers", lambda *args: MOCK_HEADERS)
     @activate(registry=registries.OrderedRegistry)
     def test_call_with_timeout_retried_to_exhaustion(self):
         post(MCD_API_ENDPOINT, body=Timeout())
@@ -276,7 +446,9 @@ class ClientTest(TestCase):
 
         backoff = MockBackoff(4)
         with self.assertRaises(expected_exception=(GqlError, Timeout)) as err:
-            self._client(query=MOCK_GET_USER_QUERY, retry_backoff=backoff, timeout_in_seconds=MOCK_TIMEOUT)
+            self._client(
+                query=MOCK_GET_USER_QUERY, retry_backoff=backoff, timeout_in_seconds=MOCK_TIMEOUT
+            )
         self.assertIsInstance(err.exception, Timeout)
         mock.assert_call_count(MCD_API_ENDPOINT, 5)
 
@@ -284,13 +456,17 @@ class ClientTest(TestCase):
         self._session.scope = _AIRFLOW_SCOPE
 
         with self.assertRaises(InvalidSessionError):
-            self._client(query=MOCK_GET_USER_QUERY, retry_backoff=self._backoff, timeout_in_seconds=MOCK_TIMEOUT)
+            self._client(
+                query=MOCK_GET_USER_QUERY,
+                retry_backoff=self._backoff,
+                timeout_in_seconds=MOCK_TIMEOUT,
+            )
 
     def test_rest_call_failed_without_scope(self):
         with self.assertRaises(InvalidSessionError):
             self._client.make_request(_AIRFLOW_CALLBACKS_PATH, body={})
 
-    @patch.object(Client, '_get_headers', lambda *args: MOCK_HEADERS)
+    @patch.object(Client, "_get_headers", lambda *args: MOCK_HEADERS)
     @activate(registry=registries.OrderedRegistry)
     def test_rest_call_airflow(self):
         self._session.scope = _AIRFLOW_SCOPE
@@ -303,7 +479,7 @@ class ClientTest(TestCase):
         response = self._client.make_request(_AIRFLOW_CALLBACKS_PATH)
         self.assertEqual(expected_response, response)
 
-    @patch.object(Client, '_get_headers', lambda *args: MOCK_HEADERS)
+    @patch.object(Client, "_get_headers", lambda *args: MOCK_HEADERS)
     @activate(registry=registries.OrderedRegistry)
     def test_rest_call_airflow_timeout_retry(self):
         self._session.scope = _AIRFLOW_SCOPE
@@ -321,7 +497,7 @@ class ClientTest(TestCase):
         )
         self.assertEqual(expected_response, response)
 
-    @patch.object(Client, '_get_headers', lambda *args: MOCK_HEADERS)
+    @patch.object(Client, "_get_headers", lambda *args: MOCK_HEADERS)
     @activate(registry=registries.OrderedRegistry)
     def test_rest_call_airflow_timeout_retries_exceeded(self):
         self._session.scope = _AIRFLOW_SCOPE
@@ -338,7 +514,7 @@ class ClientTest(TestCase):
             )
         self.assertEqual("3", str(exc.exception))
 
-    @patch.object(Client, '_get_headers', lambda *args: MOCK_HEADERS)
+    @patch.object(Client, "_get_headers", lambda *args: MOCK_HEADERS)
     @activate(registry=registries.OrderedRegistry)
     def test_rest_call_airflow_with_session(self):
         # test that we're setting the endpoint properly when initializing
